# Copyright (c) 2017 DataCore Software Corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for the base Driver for DataCore SANsymphony storage array."""

from __future__ import division

import abc
import math
from unittest import mock

from oslo_utils import units

from cinder import context
from cinder import exception as cinder_exception
from cinder.tests.unit import fake_constants
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import utils as testutils
from cinder.volume import configuration as conf
from cinder.volume.drivers.datacore import driver as datacore_driver
from cinder.volume.drivers.datacore import exception as datacore_exception
from cinder.volume.drivers.san import san


SERVER_GROUPS = [
    mock.Mock(Id='server_group_id1',
              OurGroup=True),
    mock.Mock(Id='server_group_id2',
              OurGroup=False),
]

SERVERS = [
    mock.Mock(Id='server_id1',
              State='Online'),
    mock.Mock(Id='server_id2',
              State='Online'),
]

DISK_POOLS = [
    mock.Mock(Id='disk_pool_id1',
              Caption='disk_pool1',
              ServerId='server_id1',
              PoolStatus='Running'),
    mock.Mock(Id='disk_pool_id2',
              Caption='disk_pool2',
              ServerId='server_id2',
              PoolStatus='Running'),
    mock.Mock(Id='disk_pool_id3',
              Caption='disk_pool3',
              ServerId='server_id1',
              PoolStatus='Offline'),
    mock.Mock(Id='disk_pool_id4',
              Caption='disk_pool4',
              ServerId='server_id2',
              PoolStatus='Unknown'),
]

DISK_POOL_PERFORMANCE = [
    mock.Mock(ObjectId='disk_pool_id1',
              PerformanceData=mock.Mock(BytesTotal=5 * units.Gi,
                                        BytesAllocated=2 * units.Gi,
                                        BytesAvailable=3 * units.Gi,
                                        BytesReserved=0)),
    mock.Mock(ObjectId='disk_pool_id2',
              PerformanceData=mock.Mock(BytesTotal=5 * units.Gi,
                                        BytesAllocated=3 * units.Gi,
                                        BytesAvailable=1 * units.Gi,
                                        BytesReserved=1 * units.Gi)),
    mock.Mock(ObjectId='disk_pool_id3',
              PerformanceData=None),
    mock.Mock(ObjectId='disk_pool_id4',
              PerformanceData=None),
]

STORAGE_PROFILES = [
    mock.Mock(Id='storage_profile_id1',
              Caption='storage_profile1'),
    mock.Mock(Id='storage_profile_id2',
              Caption='storage_profile2'),
    mock.Mock(Id='storage_profile_id3',
              Caption='storage_profile3'),
]

VIRTUAL_DISKS = [
    mock.Mock(Id='virtual_disk_id1',
              DiskStatus='Online',
              IsServed=False,
              Alias='virtual_disk_id1',
              Size=mock.Mock(Value=1 * units.Gi),
              FirstHostId='server_id1'),
    mock.Mock(Id='virtual_disk_id2',
              DiskStatus='Failed',
              IsServed=False,
              Alias='virtual_disk_id2',
              Size=mock.Mock(Value=1 * units.Gi),
              FirstHostId='server_id2'),
    mock.Mock(Id='virtual_disk_id3',
              DiskStatus='Online',
              IsServed=True,
              Alias='virtual_disk_id3',
              Size=mock.Mock(Value=1 * units.Gi),
              FirstHostId='server_id1',
              SecondHostId='server_id2'),
    mock.Mock(Id='virtual_disk_id4',
              DiskStatus='Failed',
              IsServed=False,
              Alias='virtual_disk_id4',
              Size=mock.Mock(Value=1 * units.Gi),
              FirstHostId='server_id1',
              SecondHostId='server_id2'),
]


EXT_VIRTUAL_DISKS = [
    mock.Mock(Id='virtual_disk_id1',
              DiskStatus='Online',
              IsServed=False,
              Alias='virtual_disk_id1',
              Size=mock.Mock(Value=2 * units.Gi),
              FirstHostId='server_id1'),
]


VIRTUAL_DISK_SNAPSHOTS = [
    mock.Mock(Id='snapshot_id1',
              State='Migrated',
              Failure='NoFailure',
              DestinationLogicalDiskId='logical_disk_id1'),
    mock.Mock(Id='snapshot_id2',
              State='Failed',
              Failure='NotAccessible',
              DestinationLogicalDiskId='logical_disk_id2'),
    mock.Mock(Id='snapshot_id3',
              State='Migrated',
              Failure='NoFailure',
              DestinationLogicalDiskId='logical_disk_id2'),
]

LOGICAL_DISKS = [
    mock.Mock(Id='logical_disk_id1',
              VirtualDiskId='virtual_disk_id1',
              ServerHostId='server_id1',
              PoolId='disk_pool_id1',
              Size=mock.Mock(Value=1 * units.Gi)),
    mock.Mock(Id='logical_disk_id2',
              VirtualDiskId='virtual_disk_id2',
              ServerHostId='server_id1',
              PoolId='disk_pool_id3',
              Size=mock.Mock(Value=1 * units.Gi)),
    mock.Mock(Id='logical_disk_id3',
              VirtualDiskId='virtual_disk_id3',
              ServerHostId='server_id1',
              PoolId='disk_pool_id1',
              Size=mock.Mock(Value=1 * units.Gi)),
    mock.Mock(Id='logical_disk_id4',
              VirtualDiskId='virtual_disk_id3',
              ServerHostId='server_id2',
              PoolId='disk_pool_id2',
              Size=mock.Mock(Value=1 * units.Gi)),
    mock.Mock(Id='logical_disk_id5',
              VirtualDiskId='virtual_disk_id4',
              ServerHostId='server_id1',
              PoolId='disk_pool_id3',
              Size=mock.Mock(Value=1 * units.Gi)),
    mock.Mock(Id='logical_disk_id6',
              VirtualDiskId='virtual_disk_id4',
              ServerHostId='server_id2',
              PoolId='disk_pool_id4',
              Size=mock.Mock(Value=1 * units.Gi)),
]

LOGICAL_UNITS = [
    mock.Mock(VirtualTargetDeviceId='target_device_id1',
              LogicalDiskId='logical_disk_id3'),
    mock.Mock(VirtualTargetDeviceId='target_device_id2',
              LogicalDiskId='logical_disk_id4'),
]

TARGET_DEVICES = [
    mock.Mock(Id='target_device_id1',
              InitiatorPortId='initiator_port_id1'),
    mock.Mock(Id='target_device_id2',
              InitiatorPortId='initiator_port_id1'),
]

CLIENTS = [
    mock.Mock(Id='client_id1',
              HostName='client_host_name1'),
    mock.Mock(Id='client_id2',
              HostName='client_host_name2'),
]


class DataCoreVolumeDriverTestCase(object):
    """Tests for the base Driver for DataCore SANsymphony storage array."""

    def setUp(self):
        super(DataCoreVolumeDriverTestCase, self).setUp()
        self.mock_client = mock.Mock()
        self.mock_client.get_servers.return_value = SERVERS
        self.mock_client.get_disk_pools.return_value = DISK_POOLS
        (self.mock_client.get_performance_by_type
         .return_value) = DISK_POOL_PERFORMANCE
        self.mock_client.get_virtual_disks.return_value = VIRTUAL_DISKS
        self.mock_client.get_storage_profiles.return_value = STORAGE_PROFILES
        self.mock_client.get_snapshots.return_value = VIRTUAL_DISK_SNAPSHOTS
        self.mock_client.get_logical_disks.return_value = LOGICAL_DISKS
        self.mock_client.get_clients.return_value = CLIENTS
        self.mock_client.get_server_groups.return_value = SERVER_GROUPS
        self.mock_object(datacore_driver.api,
                         'DataCoreClient',
                         return_value=self.mock_client)
        self.context = context.get_admin_context()

        self.volume_a = fake_volume.fake_volume_obj(
            self.context,
            **{'name': u'volume_1',
                'volume_type_id': None,
               'id': fake_constants.VOLUME_ID,
               'size': 1})
        self.volume_ext = fake_volume.fake_volume_obj(
            self.context,
            **{'name': u'volume_1',
                'volume_type_id': None,
               'id': fake_constants.VOLUME2_ID,
               'size': 2})
        self.snapshot_a = fake_snapshot.fake_snapshot_obj(
            self.context,
            **{'name': u'snapshot_1',
               'id': fake_constants.SNAPSHOT_ID,
               'size': 1})

    @staticmethod
    @abc.abstractmethod
    def init_driver(config):
        raise NotImplementedError()

    @staticmethod
    def create_configuration():
        config = conf.Configuration(None)
        config.append_config_values(san.san_opts)
        config.append_config_values(datacore_driver.datacore_opts)
        return config

    def setup_default_configuration(self):
        config = self.create_configuration()
        config.volume_backend_name = 'DataCore'
        config.san_ip = '127.0.0.1'
        config.san_login = 'dcsadmin'
        config.san_password = 'password'
        config.datacore_api_timeout = 0
        config.datacore_disk_failed_delay = 0
        return config

    def test_do_setup(self):
        config = self.setup_default_configuration()
        self.init_driver(config)

    def test_do_setup_failed(self):
        config = self.setup_default_configuration()
        config.san_ip = None
        self.assertRaises(cinder_exception.InvalidInput,
                          self.init_driver,
                          config)

        config = self.setup_default_configuration()
        config.san_login = None
        self.assertRaises(cinder_exception.InvalidInput,
                          self.init_driver,
                          config)

        config = self.setup_default_configuration()
        config.san_password = None
        self.assertRaises(cinder_exception.InvalidInput,
                          self.init_driver,
                          config)

    def test_get_volume_stats(self):
        aggregation = [(getattr(perf.PerformanceData, 'BytesTotal', 0),
                        getattr(perf.PerformanceData, 'BytesAvailable', 0),
                        getattr(perf.PerformanceData, 'BytesReserved', 0),)
                       for perf in DISK_POOL_PERFORMANCE]

        total, available, reserved = map(sum, zip(*aggregation))
        free = (available + reserved) / units.Gi
        reserved = 100.0 * reserved / total
        reserved = math.ceil(reserved)
        total /= units.Gi
        provisioned = sum(disk.Size.Value for disk in LOGICAL_DISKS)
        provisioned /= units.Gi
        ratio = 2.0

        config = self.setup_default_configuration()
        config.max_over_subscription_ratio = ratio
        driver = self.init_driver(config)
        expected_volume_stats = {
            'vendor_name': 'DataCore',
            'QoS_support': False,
            'total_capacity_gb': total,
            'free_capacity_gb': free,
            'provisioned_capacity_gb': provisioned,
            'reserved_percentage': reserved,
            'max_over_subscription_ratio': ratio,
            'thin_provisioning_support': True,
            'thick_provisioning_support': False,
            'online_extend_support': False,
            'volume_backend_name': driver.get_volume_backend_name(),
            'driver_version': driver.get_version(),
            'storage_protocol': driver.STORAGE_PROTOCOL,
        }
        volume_stats = driver.get_volume_stats(refresh=True)
        self.assertDictEqual(expected_volume_stats, volume_stats)
        volume_stats_cached = driver.get_volume_stats(refresh=False)
        self.assertEqual(volume_stats, volume_stats_cached)

    def test_create_volume(self):
        virtual_disk = VIRTUAL_DISKS[0]
        self.mock_client.create_virtual_disk_ex2.return_value = virtual_disk

        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        result = driver.create_volume(volume)
        self.assertIn('provider_location', result)
        self.assertEqual(virtual_disk.Id, result['provider_location'])

    def test_create_volume_mirrored_disk_type_specified(self):
        virtual_disk = VIRTUAL_DISKS[2]
        self.mock_client.create_virtual_disk_ex2.return_value = virtual_disk

        config = self.setup_default_configuration()
        config.datacore_disk_type = 'mirrored'
        driver = self.init_driver(config)
        volume = self.volume_a
        result = driver.create_volume(volume)
        self.assertIn('provider_location', result)
        self.assertEqual(virtual_disk.Id, result['provider_location'])

        driver = self.init_driver(self.setup_default_configuration())
        volume_type = {
            'extra_specs': {driver.DATACORE_DISK_TYPE_KEY: 'mirrored'}
        }
        get_volume_type = self.mock_object(datacore_driver.volume_types,
                                           'get_volume_type')
        get_volume_type.return_value = volume_type
        volume = self.volume_a
        volume['volume_type_id'] = 'volume_type_id'
        result = driver.create_volume(volume)
        self.assertIn('provider_location', result)
        self.assertEqual(virtual_disk.Id, result['provider_location'])

    def test_create_volume_profile_specified(self):
        virtual_disk = VIRTUAL_DISKS[0]
        self.mock_client.create_virtual_disk_ex2.return_value = virtual_disk

        config = self.setup_default_configuration()
        config.datacore_storage_profile = 'storage_profile1'
        driver = self.init_driver(config)
        volume = self.volume_a
        result = driver.create_volume(volume)
        self.assertIn('provider_location', result)
        self.assertEqual(virtual_disk.Id, result['provider_location'])

        volume_type = {
            'extra_specs': {
                driver.DATACORE_STORAGE_PROFILE_KEY: 'storage_profile2'
            }
        }
        get_volume_type = self.mock_object(datacore_driver.volume_types,
                                           'get_volume_type')
        get_volume_type.return_value = volume_type
        volume = self.volume_a
        volume['volume_type_id'] = 'volume_type_id'
        result = driver.create_volume(volume)
        self.assertIn('provider_location', result)
        self.assertEqual(virtual_disk.Id, result['provider_location'])

    def test_create_volume_pool_specified(self):
        virtual_disk = VIRTUAL_DISKS[0]
        self.mock_client.create_virtual_disk_ex2.return_value = virtual_disk

        config = self.setup_default_configuration()
        config.datacore_disk_pools = ['disk_pool1']
        driver = self.init_driver(config)
        volume = self.volume_a
        result = driver.create_volume(volume)
        self.assertIn('provider_location', result)
        self.assertEqual(virtual_disk.Id, result['provider_location'])

        volume_type = {
            'extra_specs': {driver.DATACORE_DISK_POOLS_KEY: 'disk_pool2'}
        }
        get_volume_type = self.mock_object(datacore_driver.volume_types,
                                           'get_volume_type')
        get_volume_type.return_value = volume_type
        volume = self.volume_a
        volume['volume_type_id'] = 'volume_type_id'
        result = driver.create_volume(volume)
        self.assertIn('provider_location', result)
        self.assertEqual(virtual_disk.Id, result['provider_location'])

    def test_create_volume_failed(self):
        def fail_with_datacore_fault(*args):
            raise datacore_exception.DataCoreFaultException(
                reason="General error.")

        (self.mock_client.create_virtual_disk_ex2
         .side_effect) = fail_with_datacore_fault

        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        self.assertRaises(datacore_exception.DataCoreFaultException,
                          driver.create_volume,
                          volume)

    def test_create_volume_unknown_disk_type_specified(self):
        config = self.setup_default_configuration()
        config.datacore_disk_type = 'unknown'
        driver = self.init_driver(config)
        volume = self.volume_a
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_volume,
                          volume)

        driver = self.init_driver(self.setup_default_configuration())
        volume_type = {
            'extra_specs': {driver.DATACORE_DISK_TYPE_KEY: 'unknown'}
        }
        get_volume_type = self.mock_object(datacore_driver.volume_types,
                                           'get_volume_type')
        get_volume_type.return_value = volume_type
        volume = self.volume_a
        volume['volume_type_id'] = 'volume_type_id'
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_volume,
                          volume)

    def test_create_volume_unknown_profile_specified(self):
        config = self.setup_default_configuration()
        config.datacore_storage_profile = 'unknown'
        driver = self.init_driver(config)
        volume = self.volume_a
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_volume,
                          volume)

        driver = self.init_driver(self.setup_default_configuration())
        volume_type = {
            'extra_specs': {driver.DATACORE_STORAGE_PROFILE_KEY: 'unknown'}
        }
        get_volume_type = self.mock_object(datacore_driver.volume_types,
                                           'get_volume_type')
        get_volume_type.return_value = volume_type
        volume = self.volume_a
        volume['volume_type_id'] = 'volume_type_id'
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_volume,
                          volume)

    def test_create_volume_on_failed_pool(self):
        config = self.setup_default_configuration()
        config.datacore_disk_pools = ['disk_pool3', 'disk_pool4']
        driver = self.init_driver(config)
        volume = self.volume_a
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_volume,
                          volume)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_volume_await_online_timed_out(self):
        virtual_disk = VIRTUAL_DISKS[1]
        self.mock_client.create_virtual_disk_ex2.return_value = virtual_disk

        config = self.setup_default_configuration()
        driver = self.init_driver(config)
        volume = self.volume_a
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_volume,
                          volume)

    def test_extend_volume(self):
        virtual_disk = VIRTUAL_DISKS[0]
        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        volume.provider_location = virtual_disk.Id
        self.assertIsNone(driver.extend_volume(volume, 2147483648))

    def test_extend_volume_failed_not_found(self):
        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        volume.provider_location = 'wrong_virtual_disk_id'
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.extend_volume,
                          volume,
                          2147483648)

    def test_delete_volume(self):
        virtual_disk = VIRTUAL_DISKS[0]
        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        volume.provider_location = virtual_disk.Id
        driver.delete_volume(volume)

    def test_delete_volume_assigned(self):
        self.mock_client.get_logical_disks.return_value = LOGICAL_DISKS
        self.mock_client.get_logical_units.return_value = LOGICAL_UNITS
        self.mock_client.get_target_devices.return_value = TARGET_DEVICES

        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        virtual_disk = VIRTUAL_DISKS[2]
        volume.provider_location = virtual_disk.Id
        driver.delete_volume(volume)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_snapshot(self):
        virtual_disk = VIRTUAL_DISKS[0]
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[0]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        volume.provider_location = virtual_disk.Id
        snapshot = self.snapshot_a
        snapshot.volume = volume
        result = driver.create_snapshot(snapshot)
        self.assertIn('provider_location', result)

    def test_create_snapshot_on_failed_pool(self):
        virtual_disk = VIRTUAL_DISKS[0]
        config = self.setup_default_configuration()
        config.datacore_disk_pools = ['disk_pool3', 'disk_pool4']
        driver = self.init_driver(config)
        volume = self.volume_a
        volume.provider_location = virtual_disk.Id
        snapshot = self.snapshot_a
        snapshot.volume = volume
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_snapshot,
                          snapshot)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_snapshot_await_migrated_timed_out(self):
        virtual_disk = VIRTUAL_DISKS[0]
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[1]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        volume.provider_location = virtual_disk.Id
        snapshot = self.snapshot_a
        snapshot.volume = volume
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_snapshot,
                          snapshot)

    def test_delete_snapshot(self):
        virtual_disk = VIRTUAL_DISKS[0]
        driver = self.init_driver(self.setup_default_configuration())
        snapshot = self.snapshot_a
        snapshot.provider_location = virtual_disk.Id
        driver.delete_snapshot(snapshot)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_volume_from_snapshot(self):
        virtual_disk = VIRTUAL_DISKS[0]
        self.mock_client.set_virtual_disk_size.return_value = virtual_disk
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[0]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        snapshot = self.snapshot_a
        snapshot.provider_location = virtual_disk.Id
        result = driver.create_volume_from_snapshot(volume, snapshot)
        self.assertIn('provider_location', result)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_volume_from_snapshot_mirrored_disk_type_specified(self):
        virtual_disk = VIRTUAL_DISKS[0]
        self.mock_client.set_virtual_disk_size.return_value = virtual_disk
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[0]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        config = self.setup_default_configuration()
        config.datacore_disk_type = 'mirrored'
        driver = self.init_driver(config)
        volume = self.volume_a
        snapshot = self.snapshot_a
        snapshot.provider_location = virtual_disk.Id
        result = driver.create_volume_from_snapshot(volume, snapshot)
        self.assertIn('provider_location', result)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_volume_from_snapshot_on_failed_pool(self):
        virtual_disk = VIRTUAL_DISKS[0]
        self.mock_client.set_virtual_disk_size.return_value = virtual_disk
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[0]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        config = self.setup_default_configuration()
        config.datacore_disk_type = 'mirrored'
        config.datacore_disk_pools = ['disk_pool1', 'disk_pool4']
        driver = self.init_driver(config)
        volume = self.volume_a
        snapshot = self.snapshot_a
        snapshot.provider_location = virtual_disk.Id
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_volume_from_snapshot,
                          volume,
                          snapshot)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_volume_from_snapshot_await_online_timed_out(self):
        virtual_disk = VIRTUAL_DISKS[0]
        snapshot_virtual_disk = VIRTUAL_DISKS[1]
        (self.mock_client.set_virtual_disk_size
         .return_value) = snapshot_virtual_disk
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[2]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        snapshot = self.snapshot_a
        snapshot.provider_location = virtual_disk.Id
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_volume_from_snapshot,
                          volume,
                          snapshot)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_cloned_volume(self):
        virtual_disk = VIRTUAL_DISKS[0]
        self.mock_client.set_virtual_disk_size.return_value = virtual_disk
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[0]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        src_vref = self.volume_a
        src_vref.provider_location = virtual_disk.Id
        result = driver.create_cloned_volume(volume, src_vref)
        self.assertIn('provider_location', result)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_cloned_volume_mirrored_disk_type_specified(self):
        virtual_disk = VIRTUAL_DISKS[0]
        self.mock_client.set_virtual_disk_size.return_value = virtual_disk
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[0]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        config = self.setup_default_configuration()
        config.datacore_disk_type = 'mirrored'
        driver = self.init_driver(config)
        volume = self.volume_a
        src_vref = self.volume_a
        src_vref.provider_location = virtual_disk.Id
        result = driver.create_cloned_volume(volume, src_vref)
        self.assertIn('provider_location', result)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_cloned_volume_on_failed_pool(self):
        virtual_disk = VIRTUAL_DISKS[0]
        self.mock_client.set_virtual_disk_size.return_value = virtual_disk
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[0]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        config = self.setup_default_configuration()
        config.datacore_disk_type = 'mirrored'
        config.datacore_disk_pools = ['disk_pool1', 'disk_pool4']
        driver = self.init_driver(config)
        volume = self.volume_a
        src_vref = self.volume_a
        src_vref.provider_location = virtual_disk.Id
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_cloned_volume,
                          volume,
                          src_vref)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=testutils.ZeroIntervalLoopingCall)
    def test_create_cloned_volume_await_online_timed_out(self):
        virtual_disk = VIRTUAL_DISKS[0]
        snapshot_virtual_disk = VIRTUAL_DISKS[1]
        (self.mock_client.set_virtual_disk_size
         .return_value) = snapshot_virtual_disk
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[2]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        src_vref = self.volume_a
        src_vref.provider_location = virtual_disk.Id
        self.assertRaises(cinder_exception.VolumeDriverException,
                          driver.create_cloned_volume,
                          volume,
                          src_vref)

    def test_terminate_connection(self):
        virtual_disk = VIRTUAL_DISKS[0]
        client = CLIENTS[0]
        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        volume.provider_location = virtual_disk.Id
        connector = {'host': client.HostName, 'wwpns': ['100000109bddf539']}
        driver.terminate_connection(volume, connector)

    def test_terminate_connection_connector_is_none(self):
        virtual_disk = VIRTUAL_DISKS[0]
        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_a
        volume.provider_location = virtual_disk.Id
        driver.terminate_connection(volume, None)

    def test_manage_existing(self):
        volume = self.volume_a
        driver = self.init_driver(self.setup_default_configuration())
        ret = driver.manage_existing(
            volume, self.existing_ref)
        self.assertEqual("virtual_disk_id1", ret['provider_location'])

    def test_manage_existing_get_size(self):
        volume = self.volume_a
        driver = self.init_driver(self.setup_default_configuration())
        driver.manage_existing_get_size(
            volume, self.existing_ref)

    def test_manage_existing_snapshot(self):
        snapshot = self.snapshot_a
        driver = self.init_driver(self.setup_default_configuration())
        ret = driver.manage_existing_snapshot(
            snapshot, self.existing_ref)
        self.assertEqual("virtual_disk_id1", ret['provider_location'])

    def test_manage_existing_snapshot_get_size(self):
        snapshot = self.snapshot_a
        driver = self.init_driver(self.setup_default_configuration())
        driver.manage_existing_snapshot_get_size(
            snapshot, self.existing_ref)

    def test_create_extended_cloned_volume(self):
        virtual_disk = EXT_VIRTUAL_DISKS[0]
        self.mock_client.get_virtual_disks.return_value = EXT_VIRTUAL_DISKS
        virtual_disk_snapshot = VIRTUAL_DISK_SNAPSHOTS[0]
        self.mock_client.create_snapshot.return_value = virtual_disk_snapshot

        driver = self.init_driver(self.setup_default_configuration())
        volume = self.volume_ext
        src_vref = self.volume_a
        src_vref.provider_location = virtual_disk.Id
        result = driver.create_cloned_volume(volume, src_vref)
        self.assertIn('provider_location', result)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
from oslo_utils import importutils

from cinder import context
from cinder import db
from cinder.db.sqlalchemy import api as sqla_db
from cinder.objects import fields
from cinder.objects import volume_attachment
from cinder.tests.unit.api.v2 import fakes as v2_fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.tests.unit import utils as tests_utils
from cinder.volume import configuration as conf

CONF = cfg.CONF


class AttachmentManagerTestCase(test.TestCase):
    """Attachment related test for volume.manager.py."""

    def setUp(self):
        """Setup test class."""
        super(AttachmentManagerTestCase, self).setUp()
        self.manager = importutils.import_object(CONF.volume_manager)
        self.mock_object(self.manager, '_driver_shares_targets',
                         return_value=False)
        self.configuration = mock.Mock(conf.Configuration)
        self.context = context.get_admin_context()
        self.context.user_id = fake.USER_ID
        self.project_id = fake.PROJECT3_ID
        self.context.project_id = self.project_id
        self.manager.driver.set_initialized()
        self.manager.stats = {'allocated_capacity_gb': 100,
                              'pools': {}}

    @mock.patch.object(db.sqlalchemy.api, '_volume_type_get',
                       v2_fakes.fake_volume_type_get)
    @mock.patch('cinder.db.sqlalchemy.api.volume_type_qos_specs_get')
    @mock.patch('cinder.volume.volume_types.get_volume_type_extra_specs')
    def test_attachment_update(self, get_extra_specs, mock_type_get):
        """Test attachment_update."""
        volume_params = {'status': 'available'}
        connector = {
            "initiator": "iqn.1993-08.org.debian:01:cad181614cec",
            "ip": "192.168.1.20",
            "platform": "x86_64",
            "host": "tempest-1",
            "os_type": "linux2",
            "multipath": False}

        vref = tests_utils.create_volume(self.context, **volume_params)
        self.manager.create_volume(self.context, vref)
        values = {'volume_id': vref.id,
                  'attached_host': vref.host,
                  'attach_status': 'reserved',
                  'instance_uuid': fake.UUID1,
                  'attach_mode': 'rw'}
        attachment_ref = db.volume_attach(self.context, values)
        with mock.patch.object(
                self.manager, '_notify_about_volume_usage'):
            expected = {
                'encrypted': False,
                'qos_specs': None,
                'cacheable': False,
                'access_mode': 'rw',
                'driver_volume_type': 'iscsi',
                'attachment_id': attachment_ref.id}

            get_extra_specs.return_value = ''
            self.assertEqual(expected,
                             self.manager.attachment_update(
                                 self.context,
                                 vref,
                                 connector,
                                 attachment_ref.id))
            expected = {
                'encrypted': False,
                'qos_specs': None,
                'cacheable': True,
                'access_mode': 'rw',
                'driver_volume_type': 'iscsi',
                'attachment_id': attachment_ref.id}

            get_extra_specs.return_value = '<is> True'
            self.assertEqual(expected,
                             self.manager.attachment_update(
                                 self.context,
                                 vref,
                                 connector,
                                 attachment_ref.id))

            new_attachment_ref = db.volume_attachment_get(self.context,
                                                          attachment_ref.id)
            self.assertEqual(attachment_ref.instance_uuid,
                             new_attachment_ref['instance_uuid'])
            self.assertEqual(connector['host'],
                             new_attachment_ref['attached_host'])
            self.assertEqual('na', new_attachment_ref['mountpoint'])
            self.assertEqual('rw', new_attachment_ref['attach_mode'])

            new_volume_ref = db.volume_get(self.context, vref.id)
            self.assertEqual('attaching', new_volume_ref.status)
            self.assertEqual(fields.VolumeAttachStatus.ATTACHING,
                             new_volume_ref.attach_status)

    def test_attachment_delete(self):
        """Test attachment_delete."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        self.manager.create_volume(self.context, vref)
        values = {'volume_id': vref.id,
                  'volume_host': vref.host,
                  'attach_status': 'reserved',
                  'instance_uuid': fake.UUID1}
        attachment_ref = db.volume_attach(self.context, values)
        attachment_ref = db.volume_attachment_get(
            self.context,
            attachment_ref['id'])

        vref.refresh()
        expected_status = (vref.status, vref.attach_status,
                           attachment_ref.attach_status)

        self.manager.attachment_delete(self.context,
                                       attachment_ref['id'],
                                       vref)
        # Manager doesn't change the resource status. It is changed on the API
        attachment_ref = db.volume_attachment_get(self.context,
                                                  attachment_ref.id)
        vref.refresh()
        self.assertEqual(
            expected_status,
            (vref.status, vref.attach_status, attachment_ref.attach_status))

    def test_attachment_delete_remove_export_fail(self):
        """attachment_delete removes attachment on remove_export failure."""
        self.mock_object(self.manager.driver, 'remove_export',
                         side_effect=Exception)
        # Report that the connection is not shared
        self.mock_object(self.manager, '_connection_terminate',
                         return_value=False)

        vref = tests_utils.create_volume(self.context, status='in-use',
                                         attach_status='attached')
        values = {'volume_id': vref.id, 'volume_host': vref.host,
                  'attach_status': 'reserved', 'instance_uuid': fake.UUID1}
        attach = db.volume_attach(self.context, values)
        # Confirm the volume OVO has the attachment before the deletion
        vref.refresh()
        expected_vol_status = (vref.status, vref.attach_status)
        self.assertEqual(1, len(vref.volume_attachment))

        self.manager.attachment_delete(self.context, attach.id, vref)

        # Manager doesn't change the resource status. It is changed on the API
        attachment = db.volume_attachment_get(self.context, attach.id)
        self.assertEqual(attach.attach_status, attachment.attach_status)

        vref = db.volume_get(self.context, vref.id)
        self.assertEqual(expected_vol_status,
                         (vref.status, vref.attach_status))

    def test_attachment_delete_multiple_attachments(self):
        volume_params = {'status': 'available'}
        vref = tests_utils.create_volume(self.context, **volume_params)
        attachment1 = volume_attachment.VolumeAttachment()
        attachment2 = volume_attachment.VolumeAttachment()

        attachment1.id = fake.UUID1
        attachment2.id = fake.UUID2

        @mock.patch('cinder.objects.VolumeAttachment.get_by_id',
                    side_effect=[attachment1, attachment2])
        @mock.patch.object(sqla_db, 'volume_admin_metadata_delete')
        @mock.patch.object(sqla_db, 'volume_detached')
        @mock.patch.object(self.context, 'elevated')
        @mock.patch.object(self.manager, '_connection_terminate')
        @mock.patch.object(self.manager.driver, 'remove_export')
        def _test(mock_rm_export, mock_con_term, mock_elevated,
                  mock_db_detached, mock_db_meta_delete, mock_get_attachment):
            mock_elevated.return_value = self.context
            mock_con_term.return_value = False
            mock_db_detached.return_value = (
                {'status': 'available',
                 'attach_status': fields.VolumeAttachStatus.DETACHED},
                {'attach_status': fields.VolumeAttachStatus.DETACHED,
                 'deleted': True}
            )

            # test single attachment. This should call
            # detach and remove_export
            vref.volume_attachment.objects.append(attachment1)

            self.manager.attachment_delete(self.context, attachment1.id, vref)

            mock_db_detached.called_once_with(self.context, vref,
                                              attachment1.id)
            mock_db_meta_delete.called_once_with(self.context, vref.id,
                                                 'attached_mode')
            mock_rm_export.assert_called_once_with(self.context, vref)

            # test more than 1 attachment. This should skip
            # detach and remove_export
            mock_con_term.return_value = True
            vref.volume_attachment.objects.append(attachment2)

            mock_rm_export.reset_mock()
            mock_db_detached.reset_mock()
            mock_db_meta_delete.reset_mock()

            self.manager.attachment_delete(self.context, attachment2.id, vref)

            mock_rm_export.assert_not_called()
            mock_db_detached.called_once_with(self.context, vref,
                                              attachment2.id)
            mock_db_meta_delete.called_once_with(self.context, vref.id,
                                                 'attached_mode')
        _test()

    def test_connection_terminate_no_connector_force_false(self):
        # Tests that calling _connection_terminate with an attachment that
        # does not have a connector will not call the driver and return None
        # if the force flag is False.
        attachment = mock.MagicMock(connector={})
        with mock.patch.object(self.manager.driver, '_initialized',
                               create=True, new=True):
            with mock.patch.object(self.manager.driver,
                                   'terminate_connection') as term_conn:
                has_shared_connection = self.manager._connection_terminate(
                    self.context, mock.sentinel.volume, attachment)
                self.assertIsNone(has_shared_connection)
                term_conn.assert_not_called()

    def test_connection_terminate_no_connector_force_true(self):
        # Tests that calling _connection_terminate with an attachment that
        # does not have a connector will call the driver when force is True.
        volume = mock.MagicMock()
        attachment = mock.MagicMock(connector={})
        with mock.patch.object(self.manager.driver, '_initialized',
                               create=True, new=True):
            with mock.patch.object(self.manager.driver,
                                   'terminate_connection') as term_conn:
                has_shared_connection = self.manager._connection_terminate(
                    self.context, volume, attachment, force=True)
                self.assertFalse(has_shared_connection)
                term_conn.assert_called_once_with(volume, {}, force=True)

    @mock.patch('cinder.objects.VolumeAttachment.get_by_id')
    @mock.patch('cinder.volume.manager.VolumeManager.'
                '_notify_about_volume_usage')
    @mock.patch('cinder.volume.manager.VolumeManager._connection_terminate',
                return_value=None)
    @mock.patch('cinder.db.volume_detached')
    @mock.patch('cinder.db.volume_admin_metadata_delete')
    def test_attachment_delete_none_shared_connection(self, mock_meta_del,
                                                      mock_vol_detached,
                                                      mock_conn_term,
                                                      mock_notify,
                                                      mock_get_attachment):
        # Tests that attachment_delete does not call remove_export
        # if _connection_terminate returns None indicating there is nothing
        # to consider for the export.
        volume = mock.MagicMock()
        with mock.patch.object(self.manager.driver, '_initialized',
                               create=True, new=True):
            with mock.patch.object(self.manager.driver,
                                   'remove_export') as remove_export:
                self.manager.attachment_delete(
                    self.context, mock.sentinel.attachment_id, volume)
                mock_get_attachment.assert_called_once_with(
                    self.context, mock.sentinel.attachment_id)
                remove_export.assert_not_called()



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
from unittest import mock

from cinder import exception
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.dell_emc.powermax import (
    powermax_data as tpd)
from cinder.tests.unit.volume.drivers.dell_emc.powermax import (
    powermax_fake_objects as tpfo)
from cinder.volume.drivers.dell_emc.powermax import iscsi
from cinder.volume.drivers.dell_emc.powermax import migrate
from cinder.volume.drivers.dell_emc.powermax import provision
from cinder.volume.drivers.dell_emc.powermax import rest
from cinder.volume import volume_utils


class PowerMaxMigrateTest(test.TestCase):
    def setUp(self):
        self.data = tpd.PowerMaxData()
        self.mock_object(volume_utils, 'get_max_over_subscription_ratio')
        super(PowerMaxMigrateTest, self).setUp()
        configuration = tpfo.FakeConfiguration(
            None, 'MaskingTests', 1, 1, san_ip='1.1.1.1',
            san_login='smc', powermax_array=self.data.array,
            powermax_srp='SRP_1', san_password='smc', san_api_port=8443,
            vmax_port_groups=[self.data.port_group_name_f])
        self.mock_object(rest.PowerMaxRest, '_establish_rest_session',
                         return_value=tpfo.FakeRequestsSession())
        driver = iscsi.PowerMaxISCSIDriver(configuration=configuration)
        self.driver = driver
        self.common = self.driver.common
        self.migrate = self.common.migrate

    def test_get_masking_view_component_dict_shared_format_1(self):
        """Test for get_masking_view_component_dict, legacy case 1."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-No_SLO-8970da0c-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('No_SLO', component_dict['no_slo'])
        self.assertEqual('-8970da0c', component_dict['uuid'])
        self.assertEqual('MV', component_dict['postfix'])

    def test_get_masking_view_component_dict_shared_format_2(self):
        """Test for get_masking_view_component_dict, legacy case 2."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-No_SLO-F-8970da0c-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('-F', component_dict['protocol'])
        self.assertEqual('No_SLO', component_dict['no_slo'])
        self.assertEqual('-8970da0c', component_dict['uuid'])
        self.assertEqual('MV', component_dict['postfix'])

    def test_get_masking_view_component_dict_shared_format_3(self):
        """Test for get_masking_view_component_dict, legacy case 3."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-SRP_1-Silver-NONE-74346a64-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('SRP_1', component_dict['srp'])
        self.assertEqual('Silver', component_dict['slo'])
        self.assertEqual('NONE', component_dict['workload'])
        self.assertEqual('-74346a64', component_dict['uuid'])
        self.assertEqual('MV', component_dict['postfix'])

    def test_get_masking_view_component_dict_shared_format_4(self):
        """Test for get_masking_view_component_dict, legacy case 4."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-SRP_1-Bronze-DSS-I-1b454e9f-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('SRP_1', component_dict['srp'])
        self.assertEqual('Bronze', component_dict['slo'])
        self.assertEqual('DSS', component_dict['workload'])
        self.assertEqual('-I', component_dict['protocol'])
        self.assertEqual('-1b454e9f', component_dict['uuid'])
        self.assertEqual('MV', component_dict['postfix'])

    def test_get_masking_view_component_dict_non_shared_format_5(self):
        """Test for get_masking_view_component_dict, legacy case 5."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-No_SLO-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('No_SLO', component_dict['no_slo'])
        self.assertEqual('MV', component_dict['postfix'])

    def test_get_masking_view_component_dict_non_shared_format_6(self):
        """Test for get_masking_view_component_dict, legacy case 6."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-No_SLO-F-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('No_SLO', component_dict['no_slo'])
        self.assertEqual('-F', component_dict['protocol'])
        self.assertEqual('MV', component_dict['postfix'])

    def test_get_masking_view_component_dict_non_shared_format_7(self):
        """Test for get_masking_view_component_dict, legacy case 7."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-SRP_1-Diamond-OLTP-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('SRP_1', component_dict['srp'])
        self.assertEqual('Diamond', component_dict['slo'])
        self.assertEqual('OLTP', component_dict['workload'])
        self.assertEqual('MV', component_dict['postfix'])

    def test_get_masking_view_component_dict_non_shared_format_8(self):
        """Test for get_masking_view_component_dict, legacy case 8."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-SRP_1-Gold-NONE-F-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('SRP_1', component_dict['srp'])
        self.assertEqual('Gold', component_dict['slo'])
        self.assertEqual('NONE', component_dict['workload'])
        self.assertEqual('-F', component_dict['protocol'])
        self.assertEqual('MV', component_dict['postfix'])

    def test_get_masking_view_component_dict_host_with_dashes_no_slo(
            self):
        """Test for get_masking_view_component_dict, dashes in host."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-host-with-dashes-No_SLO-I-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('host-with-dashes', component_dict['host'])
        self.assertEqual('No_SLO', component_dict['no_slo'])
        self.assertEqual('-I', component_dict['protocol'])
        self.assertEqual('MV', component_dict['postfix'])

    def test_get_masking_view_component_dict_host_with_dashes_slo(self):
        """Test for get_masking_view_component_dict, dashes and slo."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-host-with-dashes-SRP_1-Diamond-NONE-I-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('host-with-dashes', component_dict['host'])
        self.assertEqual('SRP_1', component_dict['srp'])
        self.assertEqual('Diamond', component_dict['slo'])
        self.assertEqual('NONE', component_dict['workload'])
        self.assertEqual('-I', component_dict['protocol'])
        self.assertEqual('MV', component_dict['postfix'])

    def test_get_masking_view_component_dict_replication_enabled(self):
        """Test for get_masking_view_component_dict, replication enabled."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-SRP_1-Diamond-OLTP-I-RE-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('-I', component_dict['protocol'])
        self.assertEqual('Diamond', component_dict['slo'])
        self.assertEqual('OLTP', component_dict['workload'])
        self.assertEqual('-RE', component_dict['RE'])

    def test_get_masking_view_component_dict_compression_disabled(self):
        """Test for get_masking_view_component_dict, compression disabled."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-SRP_1-Bronze-DSS_REP-I-CD-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('-I', component_dict['protocol'])
        self.assertEqual('Bronze', component_dict['slo'])
        self.assertEqual('DSS_REP', component_dict['workload'])
        self.assertEqual('-CD', component_dict['CD'])

    def test_get_masking_view_component_dict_CD_RE(self):
        """Test for get_masking_view_component_dict, CD and RE."""
        component_dict = self.migrate.get_masking_view_component_dict(
            'OS-myhost-SRP_1-Platinum-OLTP_REP-I-CD-RE-MV', 'SRP_1')
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('-I', component_dict['protocol'])
        self.assertEqual('Platinum', component_dict['slo'])
        self.assertEqual('OLTP_REP', component_dict['workload'])
        self.assertEqual('-CD', component_dict['CD'])
        self.assertEqual('-RE', component_dict['RE'])

    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_perform_migration',
                       return_value=True)
    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_get_mvs_and_sgs_from_volume',
                       return_value=(tpd.PowerMaxData.legacy_mvs,
                                     [tpd.PowerMaxData.legacy_shared_sg]))
    @mock.patch.object(migrate.PowerMaxMigrate,
                       'get_volume_host_list',
                       return_value=['myhostB'])
    def test_do_migrate_if_candidate(
            self, mock_mvs, mock_os_host, mock_migrate):
        self.assertTrue(self.migrate.do_migrate_if_candidate(
            self.data.array, self.data.srp, self.data.device_id,
            self.data.test_volume, self.data.connector))

    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_get_mvs_and_sgs_from_volume',
                       return_value=([tpd.PowerMaxData.legacy_not_shared_mv],
                                     [tpd.PowerMaxData.legacy_not_shared_sg]))
    def test_do_migrate_if_candidate_not_shared(
            self, mock_mvs):
        self.assertFalse(self.migrate.do_migrate_if_candidate(
            self.data.array, self.data.srp, self.data.device_id,
            self.data.test_volume, self.data.connector))

    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_get_mvs_and_sgs_from_volume',
                       return_value=(tpd.PowerMaxData.legacy_mvs,
                                     [tpd.PowerMaxData.legacy_shared_sg,
                                      'non_fast_sg']))
    def test_do_migrate_if_candidate_in_multiple_sgs(
            self, mock_mvs):
        self.assertFalse(self.migrate.do_migrate_if_candidate(
            self.data.array, self.data.srp, self.data.device_id,
            self.data.test_volume, self.data.connector))

    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_perform_migration',
                       return_value=True)
    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_get_mvs_and_sgs_from_volume',
                       return_value=(tpd.PowerMaxData.legacy_mvs,
                                     [tpd.PowerMaxData.legacy_shared_sg]))
    @mock.patch.object(migrate.PowerMaxMigrate,
                       'get_volume_host_list',
                       return_value=['myhostA', 'myhostB'])
    def test_dp_migrate_if_candidate_multiple_os_hosts(
            self, mock_mvs, mock_os_host, mock_migrate):
        self.assertFalse(self.migrate.do_migrate_if_candidate(
            self.data.array, self.data.srp, self.data.device_id,
            self.data.test_volume, self.data.connector))

    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_delete_staging_masking_views')
    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_get_mvs_and_sgs_from_volume',
                       side_effect=[(tpd.PowerMaxData.staging_mvs,
                                     [tpd.PowerMaxData.staging_sg]),
                                    ([tpd.PowerMaxData.staging_mv2],
                                     [tpd.PowerMaxData.staging_sg])])
    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_create_stg_masking_views',
                       return_value=tpd.PowerMaxData.staging_mvs)
    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_create_stg_storage_group_with_vol',
                       return_value=tpd.PowerMaxData.staging_sg)
    def test_perform_migration(self, mock_sg, mock_mvs, mock_new, mock_del):
        """Test to perform migration"""
        source_sg_name = 'OS-myhost-SRP_1-Diamond-OLTP-F-SG'
        mv_details_list = list()
        mv_details_list.append(self.migrate.get_masking_view_component_dict(
            'OS-myhostA-SRP_1-Diamond-OLTP-F-1b454e9f-MV', 'SRP_1'))
        mv_details_list.append(self.migrate.get_masking_view_component_dict(
            'OS-myhostB-SRP_1-Diamond-OLTP-F-8970da0c-MV', 'SRP_1'))
        self.assertTrue(self.migrate._perform_migration(
            self.data.array, self.data.device_id, mv_details_list,
            source_sg_name, 'myhostB'))

    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_create_stg_storage_group_with_vol',
                       return_value=None)
    def test_perform_migration_storage_group_fail(self, mock_sg):
        """Test to perform migration"""
        source_sg_name = 'OS-myhost-SRP_1-Diamond-OLTP-F-SG'
        mv_details_list = list()
        mv_details_list.append(self.migrate.get_masking_view_component_dict(
            'OS-myhostA-SRP_1-Diamond-OLTP-F-1b454e9f-MV', 'SRP_1'))
        mv_details_list.append(self.migrate.get_masking_view_component_dict(
            'OS-myhostB-SRP_1-Diamond-OLTP-F-8970da0c-MV', 'SRP_1'))
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.migrate._perform_migration, self.data.array,
            self.data.device_id, mv_details_list,
            source_sg_name, 'myhostB')
        with self.assertRaisesRegex(
                exception.VolumeBackendAPIException,
                'MIGRATE - Unable to create staging storage group.'):
            self.migrate._perform_migration(
                self.data.array, self.data.device_id, mv_details_list,
                source_sg_name, 'myhostB')

    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_create_stg_masking_views',
                       return_value=[])
    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_create_stg_storage_group_with_vol',
                       return_value=tpd.PowerMaxData.staging_sg)
    def test_perform_migration_masking_views_fail(self, mock_sg, mock_mvs):
        """Test to perform migration"""
        source_sg_name = 'OS-myhost-SRP_1-Diamond-OLTP-F-SG'
        mv_details_list = list()
        mv_details_list.append(self.migrate.get_masking_view_component_dict(
            'OS-myhostA-SRP_1-Diamond-OLTP-F-1b454e9f-MV', 'SRP_1'))
        mv_details_list.append(self.migrate.get_masking_view_component_dict(
            'OS-myhostB-SRP_1-Diamond-OLTP-F-8970da0c-MV', 'SRP_1'))
        with self.assertRaisesRegex(
                exception.VolumeBackendAPIException,
                'MIGRATE - Unable to create staging masking views.'):
            self.migrate._perform_migration(
                self.data.array, self.data.device_id, mv_details_list,
                source_sg_name, 'myhostB')

    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_get_mvs_and_sgs_from_volume',
                       return_value=(tpd.PowerMaxData.staging_mvs,
                                     [tpd.PowerMaxData.staging_sg,
                                      tpd.PowerMaxData.staging_sg]))
    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_create_stg_masking_views',
                       return_value=tpd.PowerMaxData.staging_mvs)
    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_create_stg_storage_group_with_vol',
                       return_value=tpd.PowerMaxData.staging_sg)
    def test_perform_migration_sg_list_len_fail(
            self, mock_sg, mock_mvs, mock_new):
        """Test to perform migration"""
        source_sg_name = 'OS-myhost-SRP_1-Diamond-OLTP-F-SG'
        mv_details_list = list()
        mv_details_list.append(self.migrate.get_masking_view_component_dict(
            'OS-myhostA-SRP_1-Diamond-OLTP-F-1b454e9f-MV', 'SRP_1'))
        mv_details_list.append(self.migrate.get_masking_view_component_dict(
            'OS-myhostB-SRP_1-Diamond-OLTP-F-8970da0c-MV', 'SRP_1'))

        exception_message = (
            r"MIGRATE - The current storage group list has 2 "
            r"members. The list is "
            r"\[\'STG-myhostB-4732de9b-98a4-4b6d-ae4b-3cafb3d34220-SG\', "
            r"\'STG-myhostB-4732de9b-98a4-4b6d-ae4b-3cafb3d34220-SG\'\]. "
            r"Will not proceed with cleanup. Please contact customer "
            r"representative.")

        with self.assertRaisesRegex(
                exception.VolumeBackendAPIException,
                exception_message):
            self.migrate._perform_migration(
                self.data.array, self.data.device_id, mv_details_list,
                source_sg_name, 'myhostB')

    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_get_mvs_and_sgs_from_volume',
                       return_value=(tpd.PowerMaxData.staging_mvs,
                                     ['not_staging_sg']))
    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_create_stg_masking_views',
                       return_value=tpd.PowerMaxData.staging_mvs)
    @mock.patch.object(migrate.PowerMaxMigrate,
                       '_create_stg_storage_group_with_vol',
                       return_value=tpd.PowerMaxData.staging_sg)
    def test_perform_migration_stg_sg_mismatch_fail(
            self, mock_sg, mock_mvs, mock_new):
        """Test to perform migration"""
        source_sg_name = 'OS-myhost-SRP_1-Diamond-OLTP-F-SG'
        mv_details_list = list()
        mv_details_list.append(self.migrate.get_masking_view_component_dict(
            'OS-myhostA-SRP_1-Diamond-OLTP-F-1b454e9f-MV', 'SRP_1'))
        mv_details_list.append(self.migrate.get_masking_view_component_dict(
            'OS-myhostB-SRP_1-Diamond-OLTP-F-8970da0c-MV', 'SRP_1'))
        with self.assertRaisesRegex(
                exception.VolumeBackendAPIException,
                'MIGRATE - The current storage group not_staging_sg does not '
                'match STG-myhostB-4732de9b-98a4-4b6d-ae4b-3cafb3d34220-SG. '
                'Will not proceed with cleanup. Please contact customer '
                'representative.'):
            self.migrate._perform_migration(
                self.data.array, self.data.device_id, mv_details_list,
                source_sg_name, 'myhostB')

    @mock.patch.object(rest.PowerMaxRest, 'delete_masking_view')
    def test_delete_staging_masking_views(self, mock_del):
        self.assertTrue(self.migrate._delete_staging_masking_views(
            self.data.array, self.data.staging_mvs, 'myhostB'))
        mock_del.assert_called_once()

    @mock.patch.object(rest.PowerMaxRest, 'delete_masking_view')
    def test_delete_staging_masking_views_no_host_match(self, mock_del):
        self.assertFalse(self.migrate._delete_staging_masking_views(
            self.data.array, self.data.staging_mvs, 'myhostC'))
        mock_del.assert_not_called()

    @mock.patch.object(rest.PowerMaxRest, 'create_masking_view')
    @mock.patch.object(rest.PowerMaxRest, 'get_masking_view',
                       return_value=tpd.PowerMaxData.maskingview[0])
    def test_create_stg_masking_views(self, mock_get, mock_create):
        mv_detail_list = list()
        for masking_view in self.data.legacy_mvs:
            masking_view_dict = self.migrate.get_masking_view_component_dict(
                masking_view, 'SRP_1')
            if masking_view_dict:
                mv_detail_list.append(masking_view_dict)
        self.assertIsNotNone(self.migrate._create_stg_masking_views(
            self.data.array, mv_detail_list, self.data.staging_sg,
            self.data.extra_specs))
        self.assertEqual(2, mock_create.call_count)

    @mock.patch.object(rest.PowerMaxRest, 'create_masking_view')
    @mock.patch.object(rest.PowerMaxRest, 'get_masking_view',
                       side_effect=[tpd.PowerMaxData.maskingview[0], None])
    def test_create_stg_masking_views_mv_not_created(
            self, mock_get, mock_create):
        mv_detail_list = list()
        for masking_view in self.data.legacy_mvs:
            masking_view_dict = self.migrate.get_masking_view_component_dict(
                masking_view, 'SRP_1')
            if masking_view_dict:
                mv_detail_list.append(masking_view_dict)
        self.assertIsNone(self.migrate._create_stg_masking_views(
            self.data.array, mv_detail_list, self.data.staging_sg,
            self.data.extra_specs))

    @mock.patch.object(provision.PowerMaxProvision, 'create_volume_from_sg')
    @mock.patch.object(provision.PowerMaxProvision, 'create_storage_group',
                       return_value=tpd.PowerMaxData.staging_mvs[0])
    def test_create_stg_storage_group_with_vol(self, mock_mv, mock_create):
        self.migrate._create_stg_storage_group_with_vol(
            self.data.array, 'myhostB', self.data.extra_specs)
        mock_create.assert_called_once()

    @mock.patch.object(provision.PowerMaxProvision, 'create_volume_from_sg')
    @mock.patch.object(provision.PowerMaxProvision, 'create_storage_group',
                       return_value=None)
    def test_create_stg_storage_group_with_vol_None(
            self, mock_mv, mock_create):
        self.assertIsNone(self.migrate._create_stg_storage_group_with_vol(
            self.data.array, 'myhostB', self.data.extra_specs))

    @mock.patch.object(rest.PowerMaxRest,
                       'get_masking_views_from_storage_group',
                       return_value=tpd.PowerMaxData.legacy_mvs)
    @mock.patch.object(rest.PowerMaxRest, 'get_storage_groups_from_volume',
                       return_value=[tpd.PowerMaxData.legacy_shared_sg])
    def test_get_mvs_and_sgs_from_volume(self, mock_sgs, mock_mvs):
        mv_list, sg_list = self.migrate._get_mvs_and_sgs_from_volume(
            self.data.array, self.data.device_id)
        mock_mvs.assert_called_once()
        self.assertEqual([self.data.legacy_shared_sg], sg_list)
        self.assertEqual(self.data.legacy_mvs, mv_list)

    @mock.patch.object(rest.PowerMaxRest,
                       'get_masking_views_from_storage_group')
    @mock.patch.object(rest.PowerMaxRest, 'get_storage_groups_from_volume',
                       return_value=list())
    def test_get_mvs_and_sgs_from_volume_empty_sg_list(
            self, mock_sgs, mock_mvs):
        mv_list, sg_list = self.migrate._get_mvs_and_sgs_from_volume(
            self.data.array, self.data.device_id)
        mock_mvs.assert_not_called()
        self.assertTrue(len(sg_list) == 0)
        self.assertTrue(len(mv_list) == 0)

    def test_get_volume_host_list(self):
        volume1 = deepcopy(self.data.test_volume)
        volume1.volume_attachment.objects = [self.data.test_volume_attachment]
        os_host_list = self.migrate.get_volume_host_list(
            volume1, self.data.connector)
        self.assertEqual('HostX', os_host_list[0])

    def test_get_volume_host_list_no_attachments(self):
        _volume_attachment = deepcopy(self.data.test_volume_attachment)
        _volume_attachment.update({'connector': None})
        volume1 = deepcopy(self.data.test_volume)
        volume1.volume_attachment.objects = [_volume_attachment]
        os_host_list = self.migrate.get_volume_host_list(
            volume1, self.data.connector)
        self.assertTrue(len(os_host_list) == 0)

    @mock.patch.object(rest.PowerMaxRest,
                       'delete_masking_view')
    @mock.patch.object(rest.PowerMaxRest,
                       'get_masking_views_from_storage_group',
                       return_value=[tpd.PowerMaxData.staging_mv1])
    @mock.patch.object(rest.PowerMaxRest,
                       'get_volumes_in_storage_group',
                       return_value=[tpd.PowerMaxData.volume_id])
    def test_cleanup_staging_objects(self, mock_vols, mock_mvs, mock_del_mv):
        self.migrate.cleanup_staging_objects(
            self.data.array, [self.data.staging_sg], self.data.extra_specs)
        mock_del_mv.assert_called_once_with(
            self.data.array, self.data.staging_mv1)

    @mock.patch.object(rest.PowerMaxRest,
                       'delete_masking_view')
    def test_cleanup_staging_objects_not_staging(self, mock_del_mv):
        self.migrate.cleanup_staging_objects(
            self.data.array, [self.data.storagegroup_name_f],
            self.data.extra_specs)
        mock_del_mv.assert_not_called()

    @mock.patch.object(rest.PowerMaxRest,
                       'get_masking_views_from_storage_group')
    @mock.patch.object(rest.PowerMaxRest,
                       'get_volumes_in_storage_group',
                       return_value=[tpd.PowerMaxData.device_id,
                                     tpd.PowerMaxData.device_id2], )
    def test_cleanup_staging_objects_multiple_vols(self, mock_vols, mock_mvs):
        self.migrate.cleanup_staging_objects(
            self.data.array, [self.data.storagegroup_name_f],
            self.data.extra_specs)
        mock_mvs.assert_not_called()



# Copyright (c) 2013 - 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json

import requests
import six

from cinder.volume.drivers.dell_emc.powerflex import driver
from cinder.volume.drivers.dell_emc.powerflex import rest_client


class PowerFlexDriver(driver.PowerFlexDriver):
    """Mock PowerFlex Driver class.

    Provides some fake configuration options
    """
    def do_setup(self, context):
        self.provisioning_type = (
            "thin" if self.configuration.san_thin_provision else "thick"
        )
        self.configuration.max_over_subscription_ratio = (
            self.configuration.powerflex_max_over_subscription_ratio
        )

    def local_path(self, volume):
        pass

    def reenable_replication(self, context, volume):
        pass

    def promote_replica(self, context, volume):
        pass

    def unmanage(self, volume):
        pass


class PowerFlexClient(rest_client.RestClient):
    """Mock PowerFlex Rest Client class.

    Provides some fake configuration options
    """

    def is_volume_creation_safe(self, _pd, _sp):
        return True


class MockHTTPSResponse(requests.Response):
    """Mock HTTP Response

    Defines the https replies from the mocked calls to do_request()
    """
    def __init__(self, content, status_code=200):
        super(MockHTTPSResponse, self).__init__()

        if isinstance(content, six.text_type):
            content = content.encode('utf-8')
        self._content = content
        self.status_code = status_code

    def json(self, **kwargs):
        if isinstance(self._content, (bytes, six.text_type)):
            return super(MockHTTPSResponse, self).json(**kwargs)

        return self._content

    @property
    def text(self):
        if not isinstance(self._content, (bytes, six.text_type)):
            return json.dumps(self._content)

        return super(MockHTTPSResponse, self).text



# Copyright (c) 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base
from cinder.objects import fields as c_fields
from cinder import utils


@base.CinderObjectRegistry.register
class Cluster(base.CinderPersistentObject, base.CinderObject,
              base.CinderComparableObject):
    """Cluster Versioned Object.

    Method get_by_id supports as additional named arguments:
        - get_services: If we want to load all services from this cluster.
        - services_summary: If we want to load num_nodes and num_down_nodes
                            fields.
        - is_up: Boolean value to filter based on the cluster's up status.
        - read_deleted: Filtering based on delete status. Default value "no".
        - Any other cluster field will be used as a filter.
    """
    # Version 1.0: Initial version
    # Version 1.1: Add replication fields
    VERSION = '1.1'
    OPTIONAL_FIELDS = ('num_hosts', 'num_down_hosts', 'services')

    # NOTE: When adding a field obj_make_compatible needs to be updated
    fields = {
        'id': fields.IntegerField(),
        'name': fields.StringField(nullable=False),
        'binary': fields.StringField(nullable=False),
        'disabled': fields.BooleanField(default=False, nullable=True),
        'disabled_reason': fields.StringField(nullable=True),
        'num_hosts': fields.IntegerField(default=0, read_only=True),
        'num_down_hosts': fields.IntegerField(default=0, read_only=True),
        'last_heartbeat': fields.DateTimeField(nullable=True, read_only=True),
        'services': fields.ObjectField('ServiceList', nullable=True,
                                       read_only=True),
        # Replication properties
        'replication_status': c_fields.ReplicationStatusField(nullable=True),
        'frozen': fields.BooleanField(default=False),
        'active_backend_id': fields.StringField(nullable=True),

        # Don't add race_preventer field, as it's a DB layer internal mechanism
        # piece to prevent races and should not be touched by other layers.
    }

    @classmethod
    def _get_expected_attrs(cls, context, *args, **kwargs):
        """Return expected attributes when getting a cluster.

        Expected attributes depend on whether we are retrieving all related
        services as well as if we are getting the services summary.
        """
        expected_attrs = []
        if kwargs.get('get_services'):
            expected_attrs.append('services')
        if kwargs.get('services_summary'):
            expected_attrs.extend(('num_hosts', 'num_down_hosts'))
        return expected_attrs

    @staticmethod
    def _from_db_object(context, cluster, db_cluster, expected_attrs=None):
        """Fill cluster OVO fields from cluster ORM instance."""
        expected_attrs = expected_attrs or tuple()
        for name, field in cluster.fields.items():
            # The only field that cannot be assigned using setattr is services,
            # because it is an ObjectField.   So we don't assign the value if
            # it's a non expected optional field or if it's services field.
            if ((name in Cluster.OPTIONAL_FIELDS
                 and name not in expected_attrs) or name == 'services'):
                continue
            value = getattr(db_cluster, name)
            setattr(cluster, name, value)

        cluster._context = context
        if 'services' in expected_attrs:
            cluster.services = base.obj_make_list(
                context,
                objects.ServiceList(context),
                objects.Service,
                db_cluster.services)

        cluster.obj_reset_changes()
        return cluster

    def obj_load_attr(self, attrname):
        """Lazy load services attribute."""
        # NOTE(geguileo): We only allow lazy loading services to raise
        # awareness of the high cost of lazy loading num_hosts and
        # num_down_hosts, so if we are going to need this information we should
        # be certain we really need it and it should loaded when retrieving the
        # data from the DB the first time we read the OVO.
        if attrname != 'services':
            raise exception.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)
        if not self._context:
            raise exception.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())

        self.services = objects.ServiceList.get_all(
            self._context, {'cluster_name': self.name})

        self.obj_reset_changes(fields=('services',))

    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))
        updates = self.cinder_obj_get_changes()
        if updates:
            for field in self.OPTIONAL_FIELDS:
                if field in updates:
                    raise exception.ObjectActionError(
                        action='create', reason=_('%s assigned') % field)

        db_cluster = db.cluster_create(self._context, updates)
        self._from_db_object(self._context, self, db_cluster)

    def save(self):
        updates = self.cinder_obj_get_changes()
        if updates:
            for field in self.OPTIONAL_FIELDS:
                if field in updates:
                    raise exception.ObjectActionError(
                        action='save', reason=_('%s changed') % field)
            db.cluster_update(self._context, self.id, updates)
            self.obj_reset_changes()

    def destroy(self):
        with self.obj_as_admin():
            updated_values = db.cluster_destroy(self._context, self.id)
        for field, value in updated_values.items():
            setattr(self, field, value)
        self.obj_reset_changes(updated_values.keys())

    @property
    def is_up(self):
        return (self.last_heartbeat and
                self.last_heartbeat >= utils.service_expired_time(True))

    def reset_service_replication(self):
        """Reset service replication flags on promotion.

        When an admin promotes a cluster, each service member requires an
        update to maintain database consistency.
        """
        actions = {
            'replication_status': 'enabled',
            'active_backend_id': None,
        }

        expectations = {
            'cluster_name': self.name,
        }

        db.conditional_update(self._context, objects.Service.model,
                              actions, expectations)


@base.CinderObjectRegistry.register
class ClusterList(base.ObjectListBase, base.CinderObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {'objects': fields.ListOfObjectsField('Cluster')}

    @classmethod
    def get_all(cls, context, is_up=None, get_services=False,
                services_summary=False, read_deleted='no', **filters):
        """Get all clusters that match the criteria.

        :param is_up: Boolean value to filter based on the cluster's up status.
        :param get_services: If we want to load all services from this cluster.
        :param services_summary: If we want to load num_nodes and
                                 num_down_nodes fields.
        :param read_deleted: Filtering based on delete status. Default value is
                             "no".
        :param filters: Field based filters in the form of key/value.
        """

        expected_attrs = Cluster._get_expected_attrs(
            context,
            get_services=get_services,
            services_summary=services_summary)

        clusters = db.cluster_get_all(context, is_up=is_up,
                                      get_services=get_services,
                                      services_summary=services_summary,
                                      read_deleted=read_deleted,
                                      **filters)
        return base.obj_make_list(context, cls(context), Cluster, clusters,
                                  expected_attrs=expected_attrs)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from http import HTTPStatus

import webob
from webob import exc

from cinder.api.openstack import wsgi
from cinder import exception
from cinder.i18n import _
from cinder import volume


class Controller(wsgi.Controller):
    """The snapshot metadata API controller for the OpenStack API."""

    def __init__(self):
        self.volume_api = volume.API()
        super(Controller, self).__init__()

    def _get_metadata(self, context, snapshot_id):
        return self._get_snapshot_and_metadata(context, snapshot_id)[1]

    def _get_snapshot_and_metadata(self, context, snapshot_id):
        # Not found exception will be handled at the wsgi level
        snapshot = self.volume_api.get_snapshot(context, snapshot_id)
        meta = self.volume_api.get_snapshot_metadata(context, snapshot)
        return snapshot, meta

    def index(self, req, snapshot_id):
        """Returns the list of metadata for a given snapshot."""
        context = req.environ['cinder.context']
        return {'metadata': self._get_metadata(context, snapshot_id)}

    def create(self, req, snapshot_id, body):
        self.assert_valid_body(body, 'metadata')
        context = req.environ['cinder.context']
        metadata = body['metadata']

        new_metadata = self._update_snapshot_metadata(context,
                                                      snapshot_id,
                                                      metadata,
                                                      delete=False)

        return {'metadata': new_metadata}

    def update(self, req, snapshot_id, id, body):
        self.assert_valid_body(body, 'meta')
        meta_item = body['meta']

        if id not in meta_item:
            expl = _('Request body and URI mismatch')
            raise exc.HTTPBadRequest(explanation=expl)

        if len(meta_item) > 1:
            expl = _('Request body contains too many items')
            raise exc.HTTPBadRequest(explanation=expl)

        context = req.environ['cinder.context']
        self._update_snapshot_metadata(context,
                                       snapshot_id,
                                       meta_item,
                                       delete=False)

        return {'meta': meta_item}

    def update_all(self, req, snapshot_id, body):
        self.assert_valid_body(body, 'metadata')
        context = req.environ['cinder.context']
        metadata = body['metadata']

        new_metadata = self._update_snapshot_metadata(context,
                                                      snapshot_id,
                                                      metadata,
                                                      delete=True)

        return {'metadata': new_metadata}

    def _update_snapshot_metadata(self, context,
                                  snapshot_id, metadata,
                                  delete=False):
        try:
            snapshot = self.volume_api.get_snapshot(context, snapshot_id)
            return self.volume_api.update_snapshot_metadata(context,
                                                            snapshot,
                                                            metadata,
                                                            delete)
        # Not found exception will be handled at the wsgi level
        except (ValueError, AttributeError):
            msg = _("Malformed request body")
            raise exc.HTTPBadRequest(explanation=msg)

        except exception.InvalidVolumeMetadata as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        except exception.InvalidVolumeMetadataSize as error:
            raise exc.HTTPRequestEntityTooLarge(explanation=error.msg)

    def show(self, req, snapshot_id, id):
        """Return a single metadata item."""
        context = req.environ['cinder.context']
        data = self._get_metadata(context, snapshot_id)

        try:
            return {'meta': {id: data[id]}}
        except KeyError:
            raise exception.SnapshotMetadataNotFound(snapshot_id=snapshot_id,
                                                     metadata_key=id)

    def delete(self, req, snapshot_id, id):
        """Deletes an existing metadata."""
        context = req.environ['cinder.context']

        snapshot, metadata = self._get_snapshot_and_metadata(context,
                                                             snapshot_id)

        if id not in metadata:
            raise exception.SnapshotMetadataNotFound(snapshot_id=snapshot_id,
                                                     metadata_key=id)

        self.volume_api.delete_snapshot_metadata(context, snapshot, id)
        return webob.Response(status_int=HTTPStatus.OK)


def create_resource():
    return wsgi.Resource(Controller())



# Copyright (c) 2022 DataCore Software Corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Fibre Channel Driver for DataCore SANsymphony storage array."""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from cinder.common import constants
from cinder import exception as cinder_exception
from cinder.i18n import _
from cinder import interface
from cinder import utils as cinder_utils
from cinder.volume import configuration
from cinder.volume.drivers.datacore import driver
from cinder.volume.drivers.datacore import exception as datacore_exception
from cinder.volume.drivers.datacore import utils as datacore_utils
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)

datacore_fc_opts = [
    cfg.ListOpt('datacore_fc_unallowed_targets',
                default=[],
                help='List of FC targets that cannot be used to attach '
                     'volume. To prevent the DataCore FibreChannel '
                     'volume driver from using some front-end targets '
                     'in volume attachment, specify this option and list '
                     'the iqn and target machine for each target as '
                     'the value, such as '
                     '<wwpns:target name>, <wwpns:target name>, '
                     '<wwpns:target name>.'),
]

CONF = cfg.CONF
CONF.register_opts(datacore_fc_opts, group=configuration.SHARED_CONF_GROUP)


@interface.volumedriver
class FibreChannelVolumeDriver(driver.DataCoreVolumeDriver):
    """DataCore SANsymphony Fibre Channel volume driver.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver
        2.0.0 - Reintroduce the driver

    """

    VERSION = '2.0.0'
    STORAGE_PROTOCOL = constants.FC
    CI_WIKI_NAME = 'DataCore_CI'

    def __init__(self, *args, **kwargs):
        super(FibreChannelVolumeDriver, self).__init__(*args, **kwargs)
        self.configuration = kwargs.get('configuration', None)
        if self.configuration:
            self.configuration.append_config_values(datacore_fc_opts)

    @classmethod
    def get_driver_options(cls):
        additional_opts = cls._get_oslo_driver_opts(
            'san_ip', 'san_login', 'san_password')
        return driver.datacore_opts + datacore_fc_opts + additional_opts

    def validate_connector(self, connector):
        """Fail if connector doesn't contain all the data needed by the driver.

        :param connector: Connector information
        """

        required_data = ['host', 'wwpns']
        for required in required_data:
            if required not in connector:
                LOG.error("The volume driver requires %(data)s "
                          "in the connector.", {'data': required})
                raise cinder_exception.InvalidConnectorException(
                    missing=required)

    def _build_initiator_target_map(self, connector):
        target_wwns = []
        init_targ_map = {}
        initiator_wwns = []

        if connector:
            initiator_wwns = connector['wwpns']
        fc_target_ports = self._get_frontend_fc_target_ports(
            self._api.get_ports())
        for target_port in fc_target_ports:
            target_wwns.append(
                target_port.PortName.replace('-', '').lower())
        for initiator in initiator_wwns:
            init_targ_map[initiator] = target_wwns

        return init_targ_map, target_wwns

    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info.

        :param volume: Volume object
        :param connector: Connector information
        :return: Connection information
        """

        LOG.debug("Initialize connection for volume %(volume)s for "
                  "connector %(connector)s.",
                  {'volume': volume.id, 'connector': connector})

        virtual_disk = self._get_virtual_disk_for(volume, raise_not_found=True)

        if virtual_disk.DiskStatus != 'Online':
            LOG.warning("Attempting to attach virtual disk %(disk)s "
                        "that is in %(state)s state.",
                        {'disk': virtual_disk.Id,
                         'state': virtual_disk.DiskStatus})

        server_group = self._get_our_server_group()

        @cinder_utils.synchronized(
            'datacore-backend-%s' % server_group.Id, external=True)
        def serve_virtual_disk():
            available_ports = self._api.get_ports()

            connector_wwpns = list(wwpn.replace('-', '').lower()
                                   for wwpn in connector['wwpns'])

            fc_initiator = self._get_initiator(connector['host'],
                                               connector_wwpns,
                                               available_ports)
            if not fc_initiator:
                msg = (_("Suitable initiator not found for "
                         "virtual disk %(disk)s for volume %(volume)s.")
                       % {'disk': virtual_disk.Id, 'volume': volume.id})
                LOG.error(msg)
                raise cinder_exception.VolumeDriverException(message=msg)

            fc_targets = self._get_targets(virtual_disk, available_ports)
            if not fc_targets:
                msg = (_("Suitable targets not found for "
                         "virtual disk %(disk)s for volume %(volume)s.")
                       % {'disk': virtual_disk.Id, 'volume': volume.id})
                LOG.error(msg)
                raise cinder_exception.VolumeDriverException(message=msg)

            virtual_logical_units = self._map_virtual_disk(
                virtual_disk, fc_targets, fc_initiator)
            return fc_targets, virtual_logical_units

        targets, logical_units = serve_virtual_disk()

        init_targ_map, target_wwns = self._build_initiator_target_map(
            connector)
        info_backend = {'driver_volume_type': 'fibre_channel',
                        'data': {
                            'target_discovered': False,
                            'target_lun': logical_units[targets[0]].Lun.Quad,
                            'target_wwn': target_wwns,
                            'volume_id': volume.id,
                            'access_mode': 'rw',
                            'initiator_target_map': init_targ_map}}

        fczm_utils.add_fc_zone(info_backend)

        LOG.debug("Connection data: %s", info_backend)

        return info_backend

    def terminate_connection(self, volume, connector, **kwargs):

        init_targ_map, target_wwns = self._build_initiator_target_map(
            connector)
        info = {'driver_volume_type': 'fibre_channel', 'data': {}}
        info['data'] = {'target_wwn': target_wwns,
                        'initiator_target_map': init_targ_map}

        # First unserve the virtual disk from Host
        super().unserve_virtual_disks_from_host(volume, connector)

        fczm_utils.remove_fc_zone(info)

        return info

    def _get_initiator(self, host, connector_wwpns, available_ports):
        wwpn_list = []
        for wwp in connector_wwpns:
            wwpn_list.append('-'.join(
                a + b for a, b in zip(*[iter(wwp.upper())] * 2)))

        client = self._get_client(host, create_new=True)
        valid_initiator = self._valid_fc_initiator(wwpn_list, available_ports)
        if not valid_initiator:
            return []

        fc_initiator_ports = self._get_host_fc_initiator_ports(
            client, available_ports)
        fc_initiator = datacore_utils.get_first_or_default(
            lambda port: True if (port.PortName in wwpn_list) else False,
            fc_initiator_ports,
            None)

        if not fc_initiator:
            for wwn in wwpn_list:
                for port in available_ports:
                    if (port.PortName == wwn and
                            port.PortType == 'FibreChannel' and
                            port.PortMode == 'Initiator' and
                            port.Connected):
                        scsi_port_data = self._api.build_scsi_port_data(
                            client.Id, wwn, 'Initiator', 'FibreChannel')
                        fc_initiator = self._api.register_port(scsi_port_data)
                        return fc_initiator
        return fc_initiator

    @staticmethod
    def _get_host_fc_initiator_ports(host, ports):
        return [port for port in ports if
                port.PortType == 'FibreChannel' and port.PortMode ==
                'Initiator' and port.HostId == host.Id]

    def _get_targets(self, virtual_disk, available_ports):
        unallowed_targets = self.configuration.datacore_fc_unallowed_targets
        fc_target_ports = self._get_frontend_fc_target_ports(
            available_ports)
        server_port_map = {}

        for target_port in fc_target_ports:
            if target_port.HostId in server_port_map:
                server_port_map[target_port.HostId].append(target_port)
            else:
                server_port_map[target_port.HostId] = [target_port]
        fc_targets = []
        if virtual_disk.FirstHostId in server_port_map:
            fc_targets += server_port_map[virtual_disk.FirstHostId]
        if virtual_disk.SecondHostId in server_port_map:
            fc_targets += server_port_map[virtual_disk.SecondHostId]

        return [target for target in fc_targets
                if target.PortName not in unallowed_targets]

    @staticmethod
    def _is_fc_frontend_port(port):
        if (port.PortType == 'FibreChannel' and
                port.PortMode == 'Target' and
                port.HostId):
            if port.PresenceStatus == 'Present':
                port_roles = port.ServerPortProperties.Role.split()
                port_state = port.StateInfo.State
                if 'Frontend' in port_roles and port_state == 'LoopLinkUp':
                    return True
        return False

    def _get_frontend_fc_target_ports(self, ports):
        return [target_port for target_port in ports
                if self._is_fc_frontend_port(target_port)]

    def _map_virtual_disk(self, virtual_disk, targets, initiator):
        logical_disks = self._api.get_logical_disks()

        logical_units = {}
        created_mapping = {}
        created_devices = []
        created_domains = []
        try:
            for target in targets:
                target_domain = self._get_target_domain(target, initiator)
                if not target_domain:
                    target_domain = self._api.create_target_domain(
                        initiator.HostId, target.HostId)
                    created_domains.append(target_domain)

                nexus = self._api.build_scsi_port_nexus_data(
                    initiator.Id, target.Id)

                target_device = self._get_target_device(
                    target_domain, target, initiator)
                if not target_device:
                    target_device = self._api.create_target_device(
                        target_domain.Id, nexus)
                    created_devices.append(target_device)

                logical_disk = self._get_logical_disk_on_host(
                    virtual_disk.Id, target.HostId, logical_disks)
                logical_unit = self._get_logical_unit(
                    logical_disk, target_device)
                if not logical_unit:
                    logical_unit = self._create_logical_unit(
                        logical_disk, nexus, target_device)
                    created_mapping[logical_unit] = target_device
                logical_units[target] = logical_unit
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Mapping operation for virtual disk %(disk)s "
                              "failed with error.",
                              {'disk': virtual_disk.Id})
                try:
                    for logical_unit in created_mapping:
                        nexus = self._api.build_scsi_port_nexus_data(
                            created_mapping[logical_unit].InitiatorPortId,
                            created_mapping[logical_unit].TargetPortId)
                        self._api.unmap_logical_disk(
                            logical_unit.LogicalDiskId, nexus)
                    for target_device in created_devices:
                        self._api.delete_target_device(target_device.Id)
                    for target_domain in created_domains:
                        self._api.delete_target_domain(target_domain.Id)
                except datacore_exception.DataCoreException as e:
                    LOG.warning("An error occurred on a cleanup after "
                                "failed mapping operation: %s.", e)

        return logical_units

    def _get_target_domain(self, target, initiator):
        target_domains = self._api.get_target_domains()
        target_domain = datacore_utils.get_first_or_default(
            lambda domain: (domain.InitiatorHostId == initiator.HostId and
                            domain.TargetHostId == target.HostId),
            target_domains, None)
        return target_domain

    def _get_target_device(self, target_domain, target, initiator):
        target_devices = self._api.get_target_devices()
        target_device = datacore_utils.get_first_or_default(
            lambda device: (device.TargetDomainId == target_domain.Id and
                            device.InitiatorPortId == initiator.Id and
                            device.TargetPortId == target.Id),
            target_devices, None)
        return target_device

    def _get_logical_unit(self, logical_disk, target_device):
        logical_units = self._api.get_logical_units()
        logical_unit = datacore_utils.get_first_or_default(
            lambda unit: (unit.LogicalDiskId == logical_disk.Id and
                          unit.VirtualTargetDeviceId == target_device.Id),
            logical_units, None)
        return logical_unit

    def _create_logical_unit(self, logical_disk, nexus, target_device):
        free_lun = self._api.get_next_free_lun(target_device.Id)
        logical_unit = self._api.map_logical_disk(logical_disk.Id,
                                                  nexus,
                                                  free_lun,
                                                  logical_disk.ServerHostId,
                                                  'Client')
        return logical_unit

    @staticmethod
    def _get_logical_disk_on_host(virtual_disk_id,
                                  host_id, logical_disks):
        logical_disk = datacore_utils.get_first(
            lambda disk: (disk.ServerHostId == host_id and
                          disk.VirtualDiskId == virtual_disk_id),
            logical_disks)
        return logical_disk

    @staticmethod
    def _valid_fc_initiator(wwpn_list, available_ports):
        for port in available_ports:
            if (port.PortType == 'FibreChannel' and
                    port.PortMode == 'Initiator'):
                if (port.PortName in wwpn_list):
                    return True
        return False



# Copyright (c) 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

STATUS_HEALTH = '1'
STATUS_ACTIVE = '43'
STATUS_RUNNING = '10'
STATUS_VOLUME_READY = '27'
STATUS_LUNCOPY_READY = '40'
STATUS_QOS_ACTIVE = '2'
QOS_INACTIVATED = '45'
LUN_TYPE = '11'
SNAPSHOT_TYPE = '27'

BLOCK_STORAGE_POOL_TYPE = '1'
FILE_SYSTEM_POOL_TYPE = '2'

HOSTGROUP_PREFIX = 'OpenStack_HostGroup_'
LUNGROUP_PREFIX = 'OpenStack_LunGroup_'
MAPPING_VIEW_PREFIX = 'OpenStack_Mapping_View_'
PORTGROUP_PREFIX = 'OpenStack_PortGroup_'
QOS_NAME_PREFIX = 'OpenStack_'
PORTGROUP_DESCRIP_PREFIX = "Please do NOT modify this. Engine ID: "
ARRAY_VERSION = 'V300R003C00'
FC_PORT_CONNECTED = '10'
FC_INIT_ONLINE = '27'
FC_PORT_MODE_FABRIC = '0'
CAPACITY_UNIT = 1024 * 1024 * 2
DEFAULT_WAIT_TIMEOUT = 3600 * 24 * 30
DEFAULT_WAIT_INTERVAL = 5

MIGRATION_WAIT_INTERVAL = 5
MIGRATION_FAULT = '74'
MIGRATION_COMPLETE = '76'

ERROR_CONNECT_TO_SERVER = -403
ERROR_UNAUTHORIZED_TO_SERVER = -401
SOCKET_TIMEOUT = 52
ERROR_VOLUME_ALREADY_EXIST = 1077948993
LOGIN_SOCKET_TIMEOUT = 4
ERROR_VOLUME_NOT_EXIST = 1077939726
RELOGIN_ERROR_PASS = [ERROR_VOLUME_NOT_EXIST]
RUNNING_NORMAL = '1'
RUNNING_SYNC = '23'
RUNNING_STOP = '41'
HEALTH_NORMAL = '1'

NO_SPLITMIRROR_LICENSE = 1077950233
NO_MIGRATION_LICENSE = 1073806606

THICK_LUNTYPE = 0
THIN_LUNTYPE = 1
MAX_NAME_LENGTH = 31
MAX_VOL_DESCRIPTION = 170
PORT_NUM_PER_CONTR = 2
PWD_EXPIRED = 3
PWD_RESET = 4

OS_TYPE = {'Linux': '0',
           'Windows': '1',
           'Solaris': '2',
           'HP-UX': '3',
           'AIX': '4',
           'XenServer': '5',
           'Mac OS X': '6',
           'VMware ESX': '7'}

HUAWEI_VALID_KEYS = ['maxIOPS', 'minIOPS', 'minBandWidth',
                     'maxBandWidth', 'latency', 'IOType']
QOS_KEYS = [i.upper() for i in HUAWEI_VALID_KEYS]
EXTRA_QOS_KEYS = ['MAXIOPS', 'MINIOPS', 'MINBANDWIDTH', 'MAXBANDWIDTH']
LOWER_LIMIT_KEYS = ['MINIOPS', 'LATENCY', 'MINBANDWIDTH']
UPPER_LIMIT_KEYS = ['MAXIOPS', 'MAXBANDWIDTH']
MAX_LUN_NUM_IN_QOS = 64

DEFAULT_REPLICA_WAIT_INTERVAL = 1
DEFAULT_REPLICA_WAIT_TIMEOUT = 20

REPLICA_SYNC_MODEL = '1'
REPLICA_ASYNC_MODEL = '2'
REPLICA_SPEED = '2'
REPLICA_PERIOD = '3600'
REPLICA_SECOND_RO = '2'
REPLICA_SECOND_RW = '3'

REPLICA_RUNNING_STATUS_KEY = 'RUNNINGSTATUS'
REPLICA_RUNNING_STATUS_INITIAL_SYNC = '21'
REPLICA_RUNNING_STATUS_SYNC = '23'
REPLICA_RUNNING_STATUS_SYNCED = '24'
REPLICA_RUNNING_STATUS_NORMAL = '1'
REPLICA_RUNNING_STATUS_SPLIT = '26'
REPLICA_RUNNING_STATUS_ERRUPTED = '34'
REPLICA_RUNNING_STATUS_INVALID = '35'

REPLICA_HEALTH_STATUS_KEY = 'HEALTHSTATUS'
REPLICA_HEALTH_STATUS_NORMAL = '1'

REPLICA_LOCAL_DATA_STATUS_KEY = 'PRIRESDATASTATUS'
REPLICA_REMOTE_DATA_STATUS_KEY = 'SECRESDATASTATUS'
REPLICA_DATA_SYNC_KEY = 'ISDATASYNC'
REPLICA_DATA_STATUS_SYNCED = '1'
REPLICA_DATA_STATUS_COMPLETE = '2'
REPLICA_DATA_STATUS_INCOMPLETE = '3'

LUN_TYPE_MAP = {'Thick': THICK_LUNTYPE,
                'Thin': THIN_LUNTYPE}

VALID_PRODUCT = ('V3', 'V5', '18000', 'Dorado')

PRODUCT_LUN_TYPE = {
    'Dorado': 'Thin',
}

VOLUME_NOT_EXISTS_WARN = 'warning'
VOLUME_NOT_EXISTS_RAISE = 'raise'

LUN_COPY_SPEED_TYPES = (
    LUN_COPY_SPEED_LOW,
    LUN_COPY_SPEED_MEDIUM,
    LUN_COPY_SPEED_HIGH,
    LUN_COPY_SPEED_HIGHEST
) = ('1', '2', '3', '4')

MAX_QUERY_COUNT = 100

CLONE_STATUS_HEALTH = '0'
CLONE_STATUS_COMPLETE = (CLONE_COMPLETE,) = ('2',)
CLONE_PAIR_NOT_EXIST = "1073798147"
SUPPORT_CLONE_PAIR_VERSION = "V600R003C00"

DEFAULT_MINIMUM_FC_INITIATOR_ONLINE = 0



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from cinder.api import microversions as mv
from cinder.tests.functional import api_samples_test_base as test_base


class SnapshotBaseTest(test_base.VolumesSampleBase):
    sample_dir = "snapshots"

    def setup(self):
        res = self._create_volume()
        res = jsonutils.loads(res.content)['volume']
        self._poll_volume_while(res['id'], ['creating'])
        self.subs = {
            "volume_id": res['id']
        }
        with self.common_api_sample():
            self.response = self._create_snapshot(self.subs)

    def _create_snapshot(self, subs=None):
        response = self._do_post('snapshots',
                                 'snapshot-create-request',
                                 subs)
        return response


@test_base.VolumesSampleBase.use_versions(
    mv.BASE_VERSION,  # 3.0
    mv.GROUP_SNAPSHOTS,  # 3.14
    mv.SNAPSHOT_LIST_USER_ID,  # 3.41
    mv.USE_QUOTA)  # 3.65
class SnapshotDetailTests(SnapshotBaseTest):
    """Test snapshot details returned for operations with different MVs.

    The details of a snapshot have changed in the different microversions, and
    we have multiple operations that return them, so we should confirm that
    each microversion returns the right values for all these different
    operations.
    """
    def test_snapshot_list_detail(self):
        response = self._do_get('snapshots/detail')
        self._verify_response('snapshots-list-detailed-response',
                              {}, response, 200)

    def test_snapshot_create(self):
        self._verify_response('snapshot-create-response',
                              {}, self.response, 202)

    def test_snapshot_show(self):
        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_get('snapshots/%s' % res['id'])
        self._verify_response('snapshot-show-response',
                              {}, response, 200)

    def test_snapshot_update(self):
        res = jsonutils.loads(self.response.content)['snapshot']
        # Use the request sample from the common API, since the request didn't
        # change with the microversion, what changes is the response.
        with self.common_api_sample():
            response = self._do_put('snapshots/%s' % res['id'],
                                    'snapshot-update-request')
        self._verify_response('snapshot-update-response',
                              {}, response, 200)


class VolumeSnapshotsSampleJsonTest(SnapshotBaseTest):
    def setUp(self):
        super(VolumeSnapshotsSampleJsonTest, self).setUp()
        self.setup()

    def test_snapshot_list(self):

        response = self._do_get('snapshots')
        self._verify_response('snapshots-list-response',
                              {}, response, 200)

    def test_snapshot_metadata_show(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_get('snapshots/%s/metadata' % res['id'])
        self._verify_response('snapshot-metadata-show-response',
                              {}, response, 200)

    def test_snapshot_metadata_create(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_post('snapshots/%s/metadata' % res['id'],
                                 'snapshot-metadata-create-request')
        self._verify_response('snapshot-metadata-create-response',
                              {}, response, 200)

    def test_snapshot_metadata_update(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_put('snapshots/%s/metadata' % res['id'],
                                'snapshot-metadata-update-request')
        self._verify_response('snapshot-metadata-update-response',
                              {}, response, 200)

    def test_snapshot_metadata_show_specific_key(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_get('snapshots/%s/metadata/key' % res['id'])
        self._verify_response('snapshot-metadata-show-key-response',
                              {}, response, 200)

    def test_snapshot_metadata_update_specific_key(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_put('snapshots/%s/metadata/key' % res['id'],
                                'snapshot-metadata-update-key-request')
        self._verify_response('snapshot-metadata-update-key-response',
                              {}, response, 200)



# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Common Auth Middleware.

"""

from http import HTTPStatus
import os

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
import webob.dec
import webob.exc

from cinder.api.openstack import wsgi
from cinder import context
from cinder.i18n import _
from cinder.wsgi import common as base_wsgi


use_forwarded_for_opt = cfg.BoolOpt(
    'use_forwarded_for',
    default=False,
    help='Treat X-Forwarded-For as the canonical remote address. '
         'Only enable this if you have a sanitizing proxy.')

CONF = cfg.CONF
CONF.register_opt(use_forwarded_for_opt)

LOG = logging.getLogger(__name__)


def pipeline_factory(loader, global_conf, **local_conf):
    """A paste pipeline replica that keys off of auth_strategy."""
    pipeline = local_conf[CONF.auth_strategy]
    if not CONF.api_rate_limit:
        limit_name = CONF.auth_strategy + '_nolimit'
        pipeline = local_conf.get(limit_name, pipeline)
    pipeline = pipeline.split()
    filters = [loader.get_filter(n) for n in pipeline[:-1]]
    app = loader.get_app(pipeline[-1])
    filters.reverse()
    for filter in filters:
        app = filter(app)
    return app


def _set_request_context(req, **kwargs):
    """Sets request context based on parameters and request."""
    remote_address = getattr(req, 'remote_addr', '127.0.0.1')

    service_catalog = None
    if req.headers.get('X_SERVICE_CATALOG') is not None:
        try:
            catalog_header = req.headers.get('X_SERVICE_CATALOG')
            service_catalog = jsonutils.loads(catalog_header)
        except ValueError:
            raise webob.exc.HTTPInternalServerError(
                explanation=_('Invalid service catalog json.'))

    if CONF.use_forwarded_for:
        remote_address = req.headers.get('X-Forwarded-For', remote_address)

    kwargs.setdefault('remote_address', remote_address)
    kwargs.setdefault('service_catalog', service_catalog)

    # Preserve the timestamp set by the RequestId middleware
    kwargs['timestamp'] = getattr(req.environ.get('cinder.context'),
                                  'timestamp',
                                  None)

    # request ID and global ID are present in the environment req.environ
    ctx = context.RequestContext.from_environ(req.environ, **kwargs)
    req.environ['cinder.context'] = ctx
    return ctx


class InjectContext(base_wsgi.Middleware):
    """Add a 'cinder.context' to WSGI environ."""

    def __init__(self, context, *args, **kwargs):
        self.context = context
        super(InjectContext, self).__init__(*args, **kwargs)

    @webob.dec.wsgify(RequestClass=base_wsgi.Request)
    def __call__(self, req):
        req.environ['cinder.context'] = self.context
        return self.application


class CinderKeystoneContext(base_wsgi.Middleware):
    """Make a request context from keystone headers."""
    ENV_OVERWRITES = {
        'X_PROJECT_DOMAIN_ID': 'project_domain_id',
        'X_PROJECT_DOMAIN_NAME': 'project_domain_name',
        'X_USER_DOMAIN_ID': 'user_domain_id',
        'X_USER_DOMAIN_NAME': 'user_domain_name',
    }

    @webob.dec.wsgify(RequestClass=base_wsgi.Request)
    def __call__(self, req):
        params = {'project_name': req.headers.get('X_TENANT_NAME')}
        for env_name, param_name in self.ENV_OVERWRITES.items():
            if req.environ.get(env_name):
                params[param_name] = req.environ[env_name]
        ctx = _set_request_context(req, **params)

        if ctx.user_id is None:
            LOG.debug("Neither X_USER_ID nor X_USER found in request")
            return webob.exc.HTTPUnauthorized()

        return self.application


class NoAuthMiddlewareBase(base_wsgi.Middleware):
    """Return a fake token if one isn't specified."""

    def base_call(self, req, project_id_in_path=False):
        if 'X-Auth-Token' not in req.headers:
            user_id = req.headers.get('X-Auth-User', 'admin')
            project_id = req.headers.get('X-Auth-Project-Id', 'admin')
            if project_id_in_path:
                os_url = os.path.join(req.url.rstrip('/'), project_id)
            else:
                os_url = req.url.rstrip('/')
            res = webob.Response()
            # NOTE(vish): This is expecting and returning Auth(1.1), whereas
            #             keystone uses 2.0 auth.  We should probably allow
            #             2.0 auth here as well.
            res.headers['X-Auth-Token'] = '%s:%s' % (user_id, project_id)
            res.headers['X-Server-Management-Url'] = os_url
            res.content_type = 'text/plain'
            res.status_int = HTTPStatus.NO_CONTENT
            return res

        token = req.headers['X-Auth-Token']
        user_id, _sep, project_id = token.partition(':')
        project_id = project_id or user_id
        _set_request_context(req, user_id=user_id, project_id=project_id,
                             is_admin=True)
        return self.application


class NoAuthMiddleware(NoAuthMiddlewareBase):
    """Return a fake token if one isn't specified.

    Sets project_id in URLs.
    """

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        return self.base_call(req)


class NoAuthMiddlewareIncludeProjectID(NoAuthMiddlewareBase):
    """Return a fake token if one isn't specified.

    Does not set project_id in URLs.
    """
    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        return self.base_call(req, project_id_in_path=True)
